@@ -14,8 +14,7 @@ from pysnmp import debug
 
 # Defaults
 verboseFlag = True
-startOID = (1, 3, 6)
-stopOID = None
+startOID = stopOID = None
 outputFile = sys.stderr
 stringPool = 'Portez ce vieux whisky au juge blond qui fume!'.split()
 int32Range = (-2147483648, 2147483648)
@@ -54,9 +53,9 @@ for opt in opts:
     if opt[0] == '--mib-module':
         modNames.append(opt[1])
     if opt[0] == '--start-oid':
-        startOID = tuple(univ.ObjectIdentifier(opt[1]))
+        startOID = univ.ObjectIdentifier(opt[1])
     if opt[0] == '--stop-oid':
-        stopOID = tuple(univ.ObjectIdentifier(opt[1]))
+        stopOID = univ.ObjectIdentifier(opt[1])
     if opt[0] == '--manual-values':
         automaticValues = False
     if opt[0] == '--output-file':
@@ -168,7 +167,12 @@ for modName in modNames:
             val = getValue(node.syntax, verboseFlag and '# Scalar %s::%s (type %s)\r\n' % (modName, symName, node.syntax.__class__.__name__) or '')
         else:
             continue
-    
+   
+        if startOID and oid < startOID:
+            continue # skip on premature OID
+        if stopOID and oid > stopOID:
+            break  # stop on out of range condition
+
         tag = '%s' % sum([ x for x in val.tagSet[0] ])
     
         # Encode non-printable content
