@@ -4,6 +4,7 @@
 # Written by Ilya Etingof <ilya@glas.net>, 2010-2012
 #
 import os
+import stat
 import sys
 import getopt
 if sys.version_info[0] < 3 and sys.version_info[1] < 5:
@@ -20,7 +21,7 @@ else:
 import bisect
 from pyasn1.type import univ
 from pyasn1.codec.ber import encoder, decoder
-from pyasn1.compat.octets import octs2str, str2octs
+from pyasn1.compat.octets import octs2str, str2octs, int2oct
 from pyasn1.error import PyAsn1Error
 from pysnmp.entity import engine, config
 from pysnmp.entity.rfc3413 import cmdrsp, context
@@ -56,6 +57,7 @@ agentUDPv4Endpoints = []
 agentUDPv6Endpoints = []
 agentUNIXEndpoints = []
 deviceDirs = set()
+interpolationModules = {}
 
 authProtocols = {
   'MD5': config.usmHMACMD5AuthProtocol,
@@ -73,11 +75,11 @@ privProtocols = {
   'NONE': config.usmNoPrivProtocol
 }
  
-helpMessage = 'Usage: %s [--help] [--debug=<category>] [--device-dir=<dir>] [--force-index-rebuild] [--validate-device-data] [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>] [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>] [--agent-unix-endpoint=</path/to/named/pipe>] [--v2c-arch] [--v3-only] [--v3-user=<username>] [--v3-auth-key=<key>] [--v3-auth-proto=<%s>] [--v3-priv-key=<key>] [--v3-priv-proto=<%s>]' % (sys.argv[0], '|'.join(authProtocols), '|'.join(privProtocols))
+helpMessage = 'Usage: %s [--help] [--debug=<category>] [--device-dir=<dir>] [--force-index-rebuild] [--validate-device-data] [--interpolation-module=</path/to/module/file[:args]>] [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>] [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>] [--agent-unix-endpoint=</path/to/named/pipe>] [--v2c-arch] [--v3-only] [--v3-user=<username>] [--v3-auth-key=<key>] [--v3-auth-proto=<%s>] [--v3-priv-key=<key>] [--v3-priv-proto=<%s>]' % (sys.argv[0], '|'.join(authProtocols), '|'.join(privProtocols))
 
 try:
     opts, params = getopt.getopt(sys.argv[1:], 'h',
-        ['help', 'debug=', 'device-dir=', 'force-index-rebuild', 'validate-device-data', 'agent-address=', 'agent-port=', 'agent-udpv4-endpoint=', 'agent-udpv6-endpoint=', 'agent-unix-endpoint=', 'v2c-arch', 'v3-only', 'v3-user=', 'v3-auth-key=', 'v3-auth-proto=', 'v3-priv-key=', 'v3-priv-proto=']
+        ['help', 'debug=', 'device-dir=', 'force-index-rebuild', 'validate-device-data', 'interpolation-module=', 'agent-address=', 'agent-port=', 'agent-udpv4-endpoint=', 'agent-udpv6-endpoint=', 'agent-unix-endpoint=', 'v2c-arch', 'v3-only', 'v3-user=', 'v3-auth-key=', 'v3-auth-proto=', 'v3-priv-key=', 'v3-priv-proto=']
         )
 except Exception:
     sys.stdout.write('%s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
@@ -99,6 +101,22 @@ for opt in opts:
         forceIndexBuild = True
     elif opt[0] == '--validate-device-data':
         validateData = True
+    elif opt[0] == '--interpolation-module':
+        args = opt[1].split(':')
+        mod, args = args[0], args[1:]
+        m = os.path.splitext(os.path.basename(mod))[0]
+        if m in interpolationModules:
+            sys.stdout.write('interpolation module %s already registered\r\n' %  m)
+            sys.exit(-1)
+        ctx = { 'path': opt[1],
+                'args': args }
+        try:
+            execfile(mod, ctx)
+        except Exception:
+            sys.stdout.write('interpolation module %s execution failure: %s\r\n' %  (opt[1], sys.exc_info()[1]))
+            sys.exit(-1)
+        else:
+            interpolationModules[m] = ctx
     elif opt[0] == '--agent-udpv4-endpoint':
         f = lambda h,p=161: (h, int(p))
         try:
@@ -181,7 +199,7 @@ class DumpParser:
         '7': rfc1902.TimeTicks,
         '8': rfc1902.Counter32,  # an alias
         '9': rfc1902.Counter64,
-        }
+    }
 
     def __nullFilter(value):
         return '' # simply drop whatever value is there when it's a Null
@@ -196,29 +214,31 @@ class DumpParser:
     filterMap = {
         '4': __nullFilter,
         '6': __unhexFilter
-        }
+    }
 
     def parse(self, line): return octs2str(line).split('|', 2)
 
     def evaluateOid(self, oid):
         return univ.ObjectIdentifier(oid)
 
-    def evaluateValue(self, tag, value):
+    def evaluateValue(self, oid, tag, value, writeMode=False, origOid=None, origValue=None, deviceFile=None):
         return tag, self.tagMap[tag](
             self.filterMap.get(tag, lambda x: x)(value.strip())
             )
     
-    def evaluate(self, line, oidOnly=False, muteValueError=True):
+    def evaluate(self, line, oidOnly=False, muteValueError=True,
+                 writeMode=False, origOid=None, origValue=None,
+                 deviceFile=None):
         oid, tag, value = self.parse(line)
         if oidOnly:
             value = None
         else:
             try:
-                tag, value = self.evaluateValue(tag, value)
+                tag, value = self.evaluateValue(oid, tag, value, writeMode, origOid, origValue)
             except:
                 if muteValueError:
                     value = rfc1902.OctetString(
-                        'SIMULATOR VALUE ERROR %s' % repr(value)
+                        'SIMULATOR VALUE ERROR %s at %s: %s' % (repr(value), deviceFile, sys.exc_info()[1])
                         )
                 else:
                     raise
@@ -227,6 +247,96 @@ class DumpParser:
 class MvcParser(DumpParser):
     ext = os.path.extsep + 'MVC'  # just an alias
 
+class SapParser(DumpParser):
+    ext = os.path.extsep + 'sapwalk'  # just an alias
+    tagMap = {
+        'Counter': rfc1902.Counter32,
+        'Gauge': rfc1902.Gauge32,
+        'Integer': rfc1902.Integer32,
+        'IpAddress': rfc1902.IpAddress,
+#        '<not implemented?>': univ.Null,
+        'ObjectID': univ.ObjectIdentifier,
+        'OctetString': rfc1902.OctetString,
+        'TimeTicks': rfc1902.TimeTicks,
+        'Counter64': rfc1902.Counter64
+    }
+
+    def __stringFilter(value):
+        if value[:2] == '0x':
+            value = [ int(value[x:x+2], 16) for x in range(2, len(value[2:])+2, 2) ]
+        return value
+
+    filterMap = {
+        'OctetString': __stringFilter
+    }
+
+    def parse(self, line):
+        return [ x.strip() for x in octs2str(line).split(',', 2) ]
+
+class WalkParser(DumpParser):
+    ext = os.path.extsep + 'snmpwalk'  # just an alias
+    # case-insensitive keys as snmpwalk output tend to vary
+    tagMap = {
+        'OID:': rfc1902.ObjectName,
+        'INTEGER:': rfc1902.Integer,
+        'STRING:': rfc1902.OctetString,
+        'BITS:': rfc1902.Bits,
+        'HEX-STRING:': rfc1902.OctetString,
+        'GAUGE32:': rfc1902.Gauge32,
+        'COUNTER32:': rfc1902.Counter32,
+        'COUNTER64:': rfc1902.Counter64,
+        'IPADDRESS:': rfc1902.IpAddress,
+        'OPAQUE:': rfc1902.Opaque,
+        'UNSIGNED32:': rfc1902.Unsigned32,  # this is not needed
+        'TIMETICKS:': rfc1902.TimeTicks     # this is made up
+    }
+
+    # possible DISPLAY-HINTs parsing should occur here
+    def __stringFilter(value):
+        if not value:
+            return value
+        elif value[0] == value[-1] == '"':
+            return value[1:-1]
+        elif value.find(':') > 0:
+            for x in value.split(':'):
+                for y in x:
+                    if y not in '0123456789ABCDEFabcdef':
+                        return value
+            return [ int(x, 16) for x in value.split(':') ]
+        else:
+            return value
+
+    def __opaqueFilter(value):
+        return [int(y, 16) for y in value.split(' ')]
+
+    def __bitsFilter(value):
+        return ''.join([int2oct(int(y, 16)) for y in value.split(' ')])
+
+    def __hexStringFilter(value):
+        return [int(y, 16) for y in value.split(' ')]
+
+    filterMap = {
+        'OPAQUE:': __opaqueFilter,
+        'STRING:': __stringFilter,
+        'BITS:': __bitsFilter,
+        'HEX-STRING:': __hexStringFilter
+    }
+
+    def parse(self, line):
+        oid, value = octs2str(line).strip().split(' = ', 1)
+        if oid and oid[0] == '.':
+            oid = oid[1:]
+        try:
+            tag, value = value.split(' ', 1)
+        except ValueError:
+            # this is implicit snmpwalk's fuzziness
+            if value == '""' or value == 'STRING:':
+                tag = 'STRING:'
+                value = ''
+            else:
+                tag = 'TimeTicks:'
+        return oid, tag.upper(), value
+
 class SnmprecParser:
     ext = os.path.extsep + 'snmprec' 
     tagMap = {}
@@ -247,24 +357,34 @@ class SnmprecParser:
     def evaluateOid(self, oid):
         return univ.ObjectIdentifier(oid)
     
-    def evaluateValue(self, tag, value):
+    def evaluateValue(self, oid, tag, value, writeMode=False, origOid=None, origValue=None, deviceFile=None):
+        # Interpolation module reference
+        if ':' in tag:
+            modName, tag = tag[tag.index(':')+1:], tag[:tag.index(':')]
+        else:
+            modName = None
         # Unhexify
         if tag[-1] == 'x':
             tag = tag[:-1]
             value = [int(value[x:x+2], 16) for x in  range(0, len(value), 2)]
+        if modName:
+            if modName in interpolationModules:
+                value = interpolationModules[modName]['interpolate'](oid, tag, value, origOid, origValue, deviceFile, writeMode)
+            else:
+                raise Exception('interpolation module "%s" missing' % modName)
         return tag, self.tagMap[tag](value)
 
-    def evaluate(self, line, oidOnly=False, muteValueError=True):
+    def evaluate(self, line, oidOnly=False, muteValueError=True, writeMode=False, origOid=None, origValue=None, deviceFile=None):
         oid, tag, value = self.parse(line)
         if oidOnly:
             value = None
         else:
             try:
-                tag, value = self.evaluateValue(tag, value)                
+                tag, value = self.evaluateValue(oid, tag, value, writeMode, origOid, origValue, deviceFile)
             except:
                 if muteValueError:
                     value = rfc1902.OctetString(
-                        'SIMULATOR VALUE ERROR %s' % repr(value)
+                        'SIMULATOR VALUE ERROR %s at %s: %s' % (repr(value), deviceFile, sys.exc_info()[1])
                         )
                 else:
                     raise
@@ -273,16 +393,23 @@ class SnmprecParser:
 parserSet = {
     DumpParser.ext: DumpParser(),
     MvcParser.ext: MvcParser(),
+    SapParser.ext: SapParser(),
+    WalkParser.ext: WalkParser(),
     SnmprecParser.ext: SnmprecParser()
-    }
+}
+
+class AbstractLayout:
+  layout = '?'
 
 # Device text file and OID index
 
-class DeviceFile:
+class DeviceFile(AbstractLayout):
+    layout = 'text'
     openedQueue = []
     maxQueueEntries = 31  # max number of open text and index files
-    def __init__(self, textFile):
+    def __init__(self, textFile, textParser):
         self.__textFile = textFile
+        self.__textParser = textParser
         try:
             self.__dbFile = textFile[:textFile.rindex(os.path.extsep)]
         except ValueError:
@@ -293,7 +420,7 @@ class DeviceFile:
         self.__db = self.__text = None
         self.__dbType = '?'
         
-    def indexText(self, textParser, forceIndexBuild=False):
+    def indexText(self, forceIndexBuild=False):
         textFileStamp = os.stat(self.__textFile)[8]
 
         # gdbm on OS X seems to voluntarily append .db, trying to catch that
@@ -347,7 +474,7 @@ class DeviceFile:
                 lineNo += 1
 
                 try:
-                    oid, tag, val = textParser.parse(line)
+                    oid, tag, val = self.__textParser.parse(line)
                 except Exception:
                     db.close()
                     exc = sys.exc_info()[1]
@@ -363,7 +490,7 @@ class DeviceFile:
 
                 if validateData:
                     try:
-                        textParser.evaluateOid(oid)
+                        self.__textParser.evaluateOid(oid)
                     except Exception:
                         db.close()
                         exc = sys.exc_info()[1]
@@ -377,15 +504,14 @@ class DeviceFile:
                                 )
                             )
                     try:
-                        textParser.evaluateValue(tag, val)
+                        self.__textParser.evaluateValue(oid, tag, val)
                     except Exception:
                         sys.stdout.write(
-                            '*** Error at line %s, value %r: %s\r\n' % \
+                            '\r\n*** Error at line %s, value %r: %s\r\n' % \
                             (lineNo, val, sys.exc_info()[1])
                             )
-                        
-                db[oid] = str(offset)
 
+                db[oid] = str(offset)
                 offset += len(line)
 
             text.close()
@@ -416,70 +542,29 @@ class DeviceFile:
 
         return self.__text, self.__db
 
-    def __str__(self):
-        return 'file %s, %s-indexed, %s' % (
-            self.__textFile, self.__dbType, self.__db and 'opened' or 'closed'
-            )
-
-# Collect device files
-
-def getDevices(tgtDir, topLen=None):
-    if topLen is None:
-        topLen = len(tgtDir.split(os.path.sep))
-    dirContent = []
-    for dFile in os.listdir(tgtDir):
-        fullPath = tgtDir + os.path.sep + dFile
-        inode = os.stat(fullPath)
-        if inode[0] & 0x4000:
-            dirContent = dirContent + getDevices(fullPath, topLen)
-            continue            
-        if not (inode[0] & 0x8000):
-            continue
-        try:
-            dExt = dFile[dFile.rindex(os.path.extsep):]
-        except ValueError:
-            continue
-        if dExt not in parserSet:
-            continue
-        fullPath = tgtDir + os.path.sep + dFile
-        relPath = fullPath.split(os.path.sep)[topLen:]
-        dirContent.append(
-            (fullPath, parserSet[dExt], os.path.sep.join(relPath)[:-len(dExt)])
-            )
-    return dirContent
-
-# Lightweignt MIB instrumentation (API-compatible with pysnmp's)
-
-class MibInstrumController:
-    eol = str2octs('\n')
-    def __init__(self, deviceFile, textParser):
-        self.__deviceFile = deviceFile
-        self.__textParser = textParser
-        self.__writeCache = {}
-
-    def __str__(self): return str(self.__deviceFile)
+    def getTextFile(self): return self.__textFile
 
     # In-place, by-OID binary search
 
-    def __searchOid(self, text, oid):
+    def __searchOid(self, oid, eol=str2octs('\n')):
         lo = mid = 0; prev_mid = -1
-        text.seek(0, 2)
-        hi = sz = text.tell()
+        self.__text.seek(0, 2)
+        hi = sz = self.__text.tell()
         while lo < hi:
             mid = (lo+hi)//2
-            text.seek(mid)
+            self.__text.seek(mid)
             while mid:
-                c = text.read(1)
-                if c == self.eol:
+                c = self.__text.read(1)
+                if c == eol:
                     mid = mid + 1
                     break
                 mid = mid - 1    # pivot stepping back in search for full line
-                text.seek(mid)
+                self.__text.seek(mid)
             if mid == prev_mid:  # loop condition due to stepping back pivot
                 break
             else:
                 prev_mid = mid
-            line = text.readline()
+            line = self.__text.readline()
             try:
                 midval, _ = self.__textParser.evaluate(line, oidOnly=True)
             except Exception:
@@ -499,7 +584,7 @@ class MibInstrumController:
         else:
             return hi
 
-    def __doVars(self, varBinds, nextFlag=False, writeMode=False):
+    def processVarBinds(self, varBinds, nextFlag=False, writeMode=False):
         rspVarBinds = []
 
         if nextFlag:
@@ -507,18 +592,19 @@ class MibInstrumController:
         else:
             errorStatus = exval.noSuchInstance
 
-        text, db = self.__deviceFile.getHandles()
+        text, db = self.getHandles()
         
         for oid, val in varBinds:
             textOid = str(
                 univ.OctetString('.'.join([ '%s' % x for x in oid ]))
                 )
+
             try:
                 offset = db[textOid]
                 exactMatch = True
             except KeyError:
                 if nextFlag:
-                    offset = self.__searchOid(text, oid)
+                    offset = self.__searchOid(oid)
                     exactMatch = False
                 else:
                     rspVarBinds.append((oid, errorStatus))
@@ -538,33 +624,72 @@ class MibInstrumController:
                 continue
 
             try:
-                _oid, _val = self.__textParser.evaluate(line)
+                _oid, _val = self.__textParser.evaluate(line, writeMode=writeMode, origOid=str(oid), origValue=str(val), deviceFile=self.getTextFile())
+            except PyAsn1Error:
+                _val = errorStatus
             except Exception:
                 raise Exception(
                     'Data error at %s for %s: %s' % (self, textOid, sys.exc_info()[1])
                     )
 
-            if writeMode:
-                try:
-                    self.__writeCache[_oid] = _val = _val.clone(val)
-                except PyAsn1Error:
-                    _val = errorStatus
-            else:
-                if _oid in self.__writeCache:
-                    _val = self.__writeCache[_oid]
-
             rspVarBinds.append((_oid, _val))
 
         return rspVarBinds
-    
+ 
+    def __str__(self):
+        return 'file %s, %s-indexed, %s' % (
+            self.__textFile, self.__dbType, self.__db and 'opened' or 'closed'
+            )
+
+# Collect device files
+
+def getDevices(tgtDir, topLen=None):
+    if topLen is None:
+        topLen = len(tgtDir.split(os.path.sep))
+    dirContent = []
+    for dFile in os.listdir(tgtDir):
+        fullPath = tgtDir + os.path.sep + dFile
+        inode = os.lstat(fullPath)
+        if stat.S_ISLNK(inode.st_mode):
+            relPath = fullPath.split(os.path.sep)[topLen:]
+            fullPath = os.readlink(fullPath)
+            if not os.path.isabs(fullPath):
+                fullPath = tgtDir + os.path.sep + fullPath
+            inode = os.stat(fullPath)
+        else:
+            relPath = fullPath.split(os.path.sep)[topLen:]
+        if stat.S_ISDIR(inode.st_mode):
+            dirContent = dirContent + getDevices(fullPath, topLen)
+            continue            
+        if not stat.S_ISREG(inode.st_mode):
+            continue
+        try:
+            dExt = dFile[dFile.rindex(os.path.extsep):]
+        except ValueError:
+            continue
+        if dExt not in parserSet:
+            continue
+        dirContent.append(
+            (fullPath, parserSet[dExt], os.path.sep.join(relPath)[:-len(dExt)])
+            )
+    return dirContent
+
+# Lightweignt MIB instrumentation (API-compatible with pysnmp's)
+
+class MibInstrumController:
+    def __init__(self, deviceFile):
+        self.__deviceFile = deviceFile
+
+    def __str__(self): return str(self.__deviceFile)
+
     def readVars(self, varBinds, acInfo=None):
-        return self.__doVars(varBinds, False)
+        return self.__deviceFile.processVarBinds(varBinds, False)
 
     def readNextVars(self, varBinds, acInfo=None):
-        return self.__doVars(varBinds, True)
+        return self.__deviceFile.processVarBinds(varBinds, True)
 
     def writeVars(self, varBinds, acInfo=None):
-        return self.__doVars(varBinds, False, True)
+        return self.__deviceFile.processVarBinds(varBinds, False, True)
 
 # Devices index as a MIB instrumentaion at a dedicated SNMP context
 
@@ -601,6 +726,10 @@ class DevicesIndexInstrumController:
 
 devicesIndexInstrumController = DevicesIndexInstrumController()
 
+mibInstrumControllerSet = {
+    DeviceFile.layout: MibInstrumController
+}
+
 # Suggest variations of context name based on request data
 def probeContext(transportDomain, transportAddress, contextName):
     candidate = [
@@ -695,20 +824,42 @@ else:
         privProtocols[v3PrivProto], v3PrivKey
         )
 
+if interpolationModules:
+    sys.stdout.write('Initializing interpolation modules:\r\n')
+    for name, body in interpolationModules.items():
+        sys.stdout.write('    %s...  ' % name)
+        for x in ('init', 'interpolate', 'shutdown'):
+            if x not in body:
+                sys.stdout.write('error: missing %s handler!\r\n' % x)
+                sys.exit(-1)
+        try:
+            body['init'](*body['args'])
+        except Exception:
+            sys.stdout.write('initialization failed: %s\r\n' % sys.exc_info()[1])
+            sys.exit(-1)
+        else:
+            sys.stdout.write('OK\r\n')
+
 # Build pysnmp Managed Objects base from device files information
 
+_mibInstrums = {}
+
 for deviceDir in deviceDirs:
     sys.stdout.write(
-        'Using data directory "%s"\r\n%s\r\n' % (deviceDir, '='*66)
+        'Scanning "%s" directory for %s data files\r\n%s\r\n' % (deviceDir, ','.join([' *%s' % x.ext for x in parserSet.values()]), '='*66)
     )
     for fullPath, textParser, communityName in getDevices(deviceDir):
-        mibInstrum = MibInstrumController(
-            DeviceFile(fullPath).indexText(textParser,
-            forceIndexBuild), textParser
-        )
+        if fullPath in _mibInstrums:
+            mibInstrum = _mibInstrums[fullPath]
+            sys.stdout.write('Shared device %s\r\n' % (mibInstrum,))
+        else:
+            deviceFile = DeviceFile(fullPath, textParser).indexText(forceIndexBuild)
+            mibInstrum = mibInstrumControllerSet[deviceFile.layout](deviceFile)
 
-        sys.stdout.write('Device %s\r\nSNMPv1/2c community name: %s\r\n' % \
-                         (mibInstrum, communityName))
+            _mibInstrums[fullPath] = mibInstrum
+            sys.stdout.write('Device %s\r\n' % (mibInstrum,))
+
+        sys.stdout.write('SNMPv1/2c community name: %s\r\n' % (communityName,))
 
         if v2cArch:
             contexts[univ.OctetString(communityName)] = mibInstrum
@@ -734,6 +885,8 @@ for deviceDir in deviceDirs:
         
         sys.stdout.write('%s\r\n' % ('-+' * 33,))
         
+del _mibInstrums
+
 if v2cArch:
     def getBulkHandler(varBinds, nonRepeaters, maxRepetitions, readNextVars):
         if nonRepeaters < 0: nonRepeaters = 0
@@ -923,6 +1076,17 @@ except KeyboardInterrupt:
 except Exception:
     exc_info = sys.exc_info()
 
+if interpolationModules:
+    sys.stdout.write('Shutting down interpolation modules:\r\n')
+    for name, body in interpolationModules.items():
+        sys.stdout.write('    %s...  ' % name)
+        try:
+            body['shutdown'](*body['args'])
+        except Exception:
+            sys.stdout.write('shutdown failed: %s\r\n' % sys.exc_info()[1])
+        else:
+            sys.stdout.write('OK\r\n')
+
 transportDispatcher.closeDispatcher()
 
 if exc_info:
